// Source file for the histogram conversion to ascii text



// Include files 

#include "RNBasics/RNBasics.h"



// Program variables

static char *histogram_name = NULL;
static float *values = NULL;
static int nvalues = 0;
static char *txt_name = NULL;
static int two_dimensional = 0;
static int print_verbose = 0;



// Global variables (only used if two_dimensional)

static int nrows = 0;
static int ncolumns = 0;



static int 
ReadHistogram(char *histogram_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = fopen(histogram_name, "rb");
  if (!fp) {
    fprintf(stderr, "Unable to open histogram file %s\n", histogram_name);
    return 0;
  }

  // Read number of values
  nvalues = 0;
  if (fread(&nvalues, sizeof(int), 1, fp) != 1) {
    fprintf(stderr, "Unable to read histogram file %s\n", histogram_name);
    return 0;
  }

  // Check if two dimensional
  if (two_dimensional) {
    // Read second number of values
    if (fread(&ncolumns, sizeof(int), 1, fp) != 1) {
      fprintf(stderr, "Unable to read histogram file %s\n", histogram_name);
      return 0;
    } 

    // Update number of values
    nrows = nvalues;
    nvalues = nrows * ncolumns;
  }

  // Allocate array of histogram values
  values = new float [ nvalues ];
  if (!values) {
    RNFail("Unable to allocate values");
    return 0;
  }

  // Read values 
  if (fread(values, sizeof(float), nvalues, fp) != (unsigned int) nvalues) {
    fprintf(stderr, "Unable to read values from histogram file %s\n", histogram_name);
    return 0;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read histogram ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Values = %d\n", nvalues);
    fflush(stdout);
  }

  // Return success
  return nvalues;
}



static int
WriteText(float *values, int nvalues, const char *txt_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = stdout;
  if (txt_name) {
    fp = fopen(txt_name, "w");
    if (!fp) {
      fprintf(stderr, "Unable to open TXT file %s\n", txt_name);
      return 0;
    }
  }

  // Write values
  for (int i = 0; i < nvalues; i++) {
    fprintf(fp, "%g\n", values[i]);
  }

  // Close file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Wrote text ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Values = %d\n", nvalues);
    fflush(stdout);
  }

  // Return number of values written
  return nvalues;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 2) {
    printf("Usage: hst2txt histogramfile [txtfile] [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) { 
        print_verbose = 1; 
      }
      if (!strcmp(*argv, "-2D")) { 
        two_dimensional = 1; 
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        exit(1); 
      }
    }
    else {
      if (!histogram_name) histogram_name = *argv;
      else if (!txt_name) txt_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        exit(1); 
      }
    }
    argv++; argc--;
  }

  // Check histogram filename
  if (!histogram_name) {
    fprintf(stderr, "You did not specify a histogram file.\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read histogram file
  int status1 = ReadHistogram(histogram_name);
  if (!status1) exit(-1);

  // Write ascii text
  int status2 = WriteText(values, nvalues, txt_name);
  if (!status2) exit(-1);

  // Return success
  return 0;
}
