// Source file for the pdb viewer program



// Include files 

#include "R3Graphics/R3Graphics.h"
#include "R3Grids/R3Grids.h"
#include "PDB/PDB.h"
#include "map.h"
#include <GL/glut.h>
#include "GL/glui.h"



// GLUI commands

enum {
  NULL_COMMAND,
  QUIT_COMMAND
};



// What a HACK ?!?!?!!?

struct MapPoint {
  MapPoint(const R3Point& position, RNScalar value) 
    : position(position), value(value) {};
  R3Point position;
  RNScalar value;
};



// Program variables

static char *map_name = NULL;
static RNScalar background_color[3] = { 1, 1, 1 };
static int make_images = 0;
static int print_verbose = 0;
static int print_debug = 0;



// GLUI variables 

static GLUI *glui = NULL;
static int GLUIwindow = 0;
static int GLUIwindow_height = 1024;
static int GLUIwindow_width = 1024;
static int GLUImouse[2] = { 0, 0 };
static int GLUIbutton[3] = { 0, 0, 0 };
static int GLUImodifiers = 0;



// Application variables

static R3Viewer *viewer = NULL;
static Map *map = NULL;
static int frame_index = 0;
static int element_index = PDB_C_ELEMENT;
static R3Point initial_camera_origin = R3Point(0.0, 0.0, 0.0);
static R3Vector initial_camera_towards = R3Vector(-0.552169, -0.708387, -0.439655);
static R3Vector initial_camera_up = R3Vector(-0.485656, 0.701917, -0.52101);
static RNBoolean initial_camera = FALSE;



// Display variables

static int show_points = 1;
static int show_box = 1;
static int show_axes = 1;
static int show_status = 1;
static RNScalar show_threshold = 1.0E-6;



void GLUIStop(void)
{
  // Destroy window 
  glutDestroyWindow(GLUIwindow);

  // Exit
  exit(0);
}



void GLUIRedraw(void)
{
  // Set viewing transformation
  viewer->Camera().Load();

  // Clear window 
  glClearColor(background_color[0], background_color[1], background_color[2], 1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Set lights
  static GLfloat light0_position[] = { 3.0, 4.0, 5.0, 0.0 };
  glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
  static GLfloat light1_position[] = { -3.0, -2.0, -3.0, 0.0 };
  glLightfv(GL_LIGHT1, GL_POSITION, light1_position);

  // Compute sum of values in map(frame_index, element_index)
  RNScalar sum = map->L1Norm(frame_index, element_index);

  // Draw map
  if (show_points) {
    if (map->storage_type == POINT_MAP_STORAGE) {
      RNArray<MapPoint *> *points = (RNArray<MapPoint *> *) map->distributions[frame_index][element_index];
      if (points) {
        glPointSize(10.0);
        glDisable(GL_LIGHTING);
        glDepthMask(FALSE);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glEnable(GL_BLEND);
        glDisable(GL_DEPTH_TEST);
        glBegin(GL_POINTS);
        for (int i = 0; i < points->NEntries(); i++) {
          RNScalar value = points->Kth(i)->value;
          const R3Point& position = points->Kth(i)->position;
          RNScalar graylevel = 0.5 * value * points->NEntries() / sum;
          glColor3d(graylevel, graylevel, graylevel);
          R3LoadPoint(position);
        }
        glEnd();
        glEnable(GL_DEPTH_TEST);
        glDisable(GL_BLEND);
        glBlendFunc(GL_ONE, GL_ZERO);
        glDepthMask(TRUE);
        glPointSize(1.0);
      }
    }
    else if (map->storage_type == GRID_MAP_STORAGE) {
      R3Grid *grid = (R3Grid *) map->distributions[frame_index][element_index];
      if (grid) {
        // Push grid transformation
        grid->GridToWorldTransformation().Push();

        // Draw grid isosurface
        glDisable(GL_LIGHTING);
        RNLoadRgb(1.0, 0.0, 0.0);
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        grid->DrawIsoSurface(show_threshold);
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

        // Pop grid transformation
        grid->GridToWorldTransformation().Pop();
      }
    }
    else if (map->storage_type == DISTANCE_MAP_STORAGE) {
      RNScalar *histogram = (RNScalar *) map->distributions[frame_index][element_index];
      if (histogram) {
        RNScalar xmax = map->Resolution()-1;
        RNScalar ymax = 0;
        for (int i = 0; i < map->res; i++) {
          if (histogram[i] > ymax) ymax = histogram[i];
        }
        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glLoadIdentity(); 
        gluOrtho2D(-0.1 * xmax, 1.1 * xmax, -0.1 * ymax, 1.1 * ymax);
        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();
        glLoadIdentity();
        glDisable(GL_LIGHTING);
        glDisable(GL_DEPTH_TEST);
        glDepthMask(FALSE);
        glLineWidth(5);
        glColor3f(0, 0, 0);
        R2BeginLine();
        R2LoadPoint(0.0, ymax);
        R2LoadPoint(0.0, 0.0);
        R2LoadPoint(xmax, 0.0);
        R2EndLine();
        char buffer[128];
        sprintf(buffer, "%g", map->WorldRadius());
        R2DrawText(xmax, 0.0, buffer);
        sprintf(buffer, "%g", ymax);
        R2DrawText(0.0, ymax, buffer);
        glColor3f(1, 0, 0);
        R2BeginLine();
        for (int i = 0; i < map->res; i++) {
          RNCoord x = i;
          RNCoord y = histogram[i];
          R2LoadPoint(x, y);
        }
        R2EndLine();
        glLineWidth(1);
        glEnable(GL_DEPTH_TEST);
        glDepthMask(TRUE);
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        glMatrixMode(GL_MODELVIEW);
        glPopMatrix();
      }
    }
  }

  // Draw map bounding box
  if (show_box) {
    if (map->storage_type != DISTANCE_MAP_STORAGE) {
      glDisable(GL_LIGHTING);
      glColor3f(0, 0, 0);
      R3Point corner(map->WorldRadius(), map->WorldRadius(), map->WorldRadius());
      R3Box(-corner, corner).Outline();
    }
  }

  // Draw map axes
  if (show_axes) {
    if (map->storage_type != DISTANCE_MAP_STORAGE) {
      RNScalar r = map->WorldRadius();
      glDisable(GL_LIGHTING);
      glLineWidth(3);
      glColor3f(1, 0, 0);
      R3BeginLine();
      glVertex3d(0, 0, 0);
      glVertex3d(r, 0, 0);
      R3EndLine();
      glColor3f(0, 1, 0);
      R3BeginLine();
      glVertex3d(0, 0, 0);
      glVertex3d(0, r, 0);
      R3EndLine();
      glColor3f(0, 0, 1);
      R3BeginLine();
      glVertex3d(0, 0, 0);
      glVertex3d(0, 0, r);
      R3EndLine();
      glLineWidth(1);
    }
  }

  // Draw status text
  if (show_status) {
    glDisable(GL_LIGHTING);
    char buffer[128];
    sprintf(buffer, "%s  ,  %s    :    %g", map->FrameName(frame_index), PDBelements[element_index].Name(), sum);
    R3Point p = viewer->WorldRay(100, 100).Point(2 * viewer->Camera().Near());
    glColor3f(1 - background_color[0], 1 - background_color[1], 1 - background_color[2]);
    R3DrawText(p, buffer);
  }

  // Capture image and exit
  if (make_images) {
    if (sum != 0) {
      char image_name[256];
      sprintf(image_name, "%s_%s_%s.jpg", map_name, map->FrameName(frame_index), PDBelements[element_index].Name());
      R2Image image(GLUIwindow_width, GLUIwindow_height, 3);
      image.Capture();
      image.Write(image_name);
    }

    // Update frame index and element_index 
    element_index++;
    if (element_index >= map->NElements()) {
      frame_index++;
      element_index = 0;
    }
    if (frame_index >= map->NFrames()) {
      frame_index = 0;
      GLUIStop();
    }
  }

  // Swap buffers 
  glutSwapBuffers();
}    



void GLUIResize(int a, int b)
{
  // Resize window
  int x, y, w, h;
  GLUI_Master.get_viewport_area( &x, &y, &w, &h );
  glViewport( x, y, w, h );

  // Resize viewer viewport
  viewer->ResizeViewport(x, y, w, h);

  // Remember window size 
  GLUIwindow_width = w;
  GLUIwindow_height = h;

  // Redraw
  glutPostRedisplay();
}



void GLUIMotion(int x, int y)
{
  // Invert y coordinate
  y = GLUIwindow_height - y;

  // Compute mouse movement
  int dx = x - GLUImouse[0];
  int dy = y - GLUImouse[1];
  
  // World in hand navigation 
  R3Point origin = R3zero_point;
  if (GLUIbutton[0]) viewer->RotateWorld(1.0, origin, x, y, dx, dy);
  else if (GLUIbutton[1]) viewer->ScaleWorld(1.0, origin, x, y, dx, dy);
  else if (GLUIbutton[2]) viewer->TranslateWorld(1.0, origin, x, y, dx, dy);
  if (GLUIbutton[0] || GLUIbutton[1] || GLUIbutton[2]) glutPostRedisplay();

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = y;
}



void GLUIMouse(int button, int state, int x, int y)
{
  // Invert y coordinate
  y = GLUIwindow_height - y;
  
  // Process mouse button event

  // Remember button state 
  int b = (button == GLUT_LEFT_BUTTON) ? 0 : ((button == GLUT_MIDDLE_BUTTON) ? 1 : 2);
  GLUIbutton[b] = (state == GLUT_DOWN) ? 1 : 0;

  // Remember modifiers 
  GLUImodifiers = glutGetModifiers();

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = y;

  // Redraw
  glutPostRedisplay();
}



void GLUISpecial(int key, int x, int y)
{
  // Invert y coordinate
  y = GLUIwindow_height - y;

  // Process keyboard button event 
  switch (key) {
  case GLUT_KEY_DOWN:
    show_threshold *= 0.9;
    if (show_threshold < 1.0E-6) show_threshold = 0.0;
    break;

  case GLUT_KEY_UP:
    show_threshold *= 1.1;
    if (show_threshold < 1.0E-6) show_threshold = 1.0E-6;
    break;
  }

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = y;

  // Remember modifiers 
  GLUImodifiers = glutGetModifiers();

  // Synchronize GLUI interface
  GLUI_Master.sync_live_all();

  // Redraw
  glutPostRedisplay();
}



void GLUIKeyboard(unsigned char key, int x, int y)
{
  // Process keyboard button event 
  switch (key) {
  case 'A':
  case 'a':
    show_axes = !show_axes;
    break;

  case 'B':
  case 'b':
    show_box = !show_box;
    break;

  case 'P':
  case 'p':
    show_points = !show_points;
    break;

  case 'S':
  case 's':
    show_status = !show_status;
    break;

  case ',':
    frame_index--;
    if (frame_index < 0) frame_index = map->NFrames() - 1;
    break;

  case '.':
    frame_index++;
    if (frame_index >= map->NFrames()) frame_index = 0;
    break;

  case '<':
    element_index--;
    if (element_index < 0) element_index = map->NElements() - 1;
    break;

  case '>':
    element_index++;
    if (element_index >= map->NElements()) element_index = 0;
    break;

  case ' ': {
    // Print camera
    const R3Camera& camera = viewer->Camera();
    printf("#camera  %g %g %g  %g %g %g  %g %g %g  %g \n",
           camera.Origin().X(), camera.Origin().Y(), camera.Origin().Z(),
           camera.Towards().X(), camera.Towards().Y(), camera.Towards().Z(),
           camera.Up().X(), camera.Up().Y(), camera.Up().Z(),
           camera.YFOV());
    break; }
      
  case 27: // ESCAPE
    GLUIStop();
    break;
  }

  // Remember mouse position 
  GLUImouse[0] = x;
  GLUImouse[1] = GLUIwindow_height - y;

  // Remember modifiers 
  GLUImodifiers = glutGetModifiers();

  // Synchronize GLUI interface
  GLUI_Master.sync_live_all();

  // Redraw
  glutPostRedisplay();  
}




void GLUICallback(int control)
{
  switch (control) {
  case QUIT_COMMAND:
    GLUIStop();
    break;
  }
}



void GLUIIdle(void)
{
  // Just checking
  assert(make_images);

  // Set current window
  if ( glutGetWindow() != GLUIwindow ) 
    glutSetWindow(GLUIwindow); 

  // Redraw
  glutPostRedisplay();
}



void GLUISetup(void)
{
  GLUI_Panel *panel;
  GLUI_Spinner *spinner;

  // Initialize GLUI
  glui = GLUI_Master.create_glui("MBE Controls", 0, GLUIwindow_width + 120, 100);
  glui->set_main_gfx_window(GLUIwindow);

  // Add frame/element selections
  panel = glui->add_panel("Map Selection");
  spinner = glui->add_spinner_to_panel(panel, "Frame", GLUI_SPINNER_INT, &frame_index);
  spinner->set_int_limits(0, map->NFrames()-1, GLUI_LIMIT_WRAP);
  spinner = glui->add_spinner_to_panel(panel, "Element", GLUI_SPINNER_INT, &element_index);
  spinner->set_int_limits(0, map->NElements()-1, GLUI_LIMIT_WRAP);

  // Add button
  glui->add_button("Quit", QUIT_COMMAND, GLUICallback);
}



void GLUIInit(int *argc, char **argv)
{
  // Open window 
  glutInit(argc, argv);
  glutInitWindowPosition(100, 100);
  glutInitWindowSize(GLUIwindow_width, GLUIwindow_height);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH); // | GLUT_STENCIL
  GLUIwindow = glutCreateWindow("OpenGL Viewer");

  // Initialize background color 
  glClearColor(200.0/255.0, 200.0/255.0, 200.0/255.0, 1.0);

  // Initialize lights
  static GLfloat lmodel_ambient[] = { 0.2, 0.2, 0.2, 1.0 };
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
  static GLfloat light0_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
  glEnable(GL_LIGHT0);
  static GLfloat light1_diffuse[] = { 0.5, 0.5, 0.5, 1.0 };
  glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
  glEnable(GL_LIGHT1);
  glEnable(GL_NORMALIZE);
  glEnable(GL_LIGHTING);

  // Initialize graphics modes 
  glEnable(GL_DEPTH_TEST);

  // Initialize GLUT callback functions 
  glutDisplayFunc(GLUIRedraw);
  GLUI_Master.set_glutReshapeFunc(GLUIResize);
  GLUI_Master.set_glutKeyboardFunc(GLUIKeyboard);
  GLUI_Master.set_glutSpecialFunc(GLUISpecial);
  GLUI_Master.set_glutMouseFunc(GLUIMouse);
  glutMotionFunc(GLUIMotion);

  // Initialize font
#if (RN_OS == RN_WINDOWSNT)
  int font = glGenLists(256);
  wglUseFontBitmaps(wglGetCurrentDC(), 0, 256, font); 
  glListBase(font);
#endif
}



void GLUIMainLoop(void)
{
  // Setup GLUI controls
  GLUISetup();

  // Idle if making images
  if (make_images) {
    frame_index = 0;
    element_index = 0;
    GLUI_Master.set_glutIdleFunc(GLUIIdle);
  }

  // Run main loop -- never returns 
  glutMainLoop();
}



static Map *
ReadMap(const char *map_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate map
  Map *map = new Map();
  if (!map) {
    fprintf(stderr, "Unable to allocate map.\n");
    return NULL;
  }

  // Read map
  int status = map->ReadFile(map_name);
  if (!status) return 0;

  // Print statistics
  if (print_verbose) {
    printf("Read map ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Frame Type = %d\n", map->FrameType());
    printf("  Storage Type = %d\n", map->StorageType());
    printf("  # Frames = %d\n", map->NFrames());
    printf("  # Elements = %d\n", map->NElements());
    printf("  World Radius = %g\n", map->WorldRadius());
    printf("  World Spacing = %g\n", map->WorldSpacing());
    printf("  Resolution = %d\n", map->Resolution());
    fflush(stdout);
  }

  // Print more statistics
  if (print_debug) map->PrintStatistics();

  // Return map
  return map;
}



static R3Viewer *
CreateViewer(Map *map)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Get pdb bounding box
  R3Point corner(map->WorldRadius(), map->WorldRadius(), map->WorldRadius());
  R3Box bbox(-corner, corner);
  assert(!bbox.IsEmpty());
  RNLength r = bbox.DiagonalRadius();
  assert((r > 0.0) && RNIsFinite(r));

  // Setup camera view looking down the Z axis
  if (!initial_camera) initial_camera_origin = bbox.Centroid() - initial_camera_towards * (2.5 * r);;
  R3Camera camera(initial_camera_origin, initial_camera_towards, initial_camera_up, 0.4, 0.4, 0.1 * r, 1000.0 * r);
  R2Viewport viewport(0, 0, GLUIwindow_width, GLUIwindow_height);
  R3Viewer *viewer = new R3Viewer(camera, viewport);

  // Print statistics
  if (print_verbose) {
    printf("Created viewer ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Origin = %g %g %g\n", camera.Origin().X(), camera.Origin().Y(), camera.Origin().Z());
    printf("  Towards = %g %g %g\n", camera.Towards().X(), camera.Towards().Y(), camera.Towards().Z());
    printf("  Up = %g %g %g\n", camera.Up().X(), camera.Up().Y(), camera.Up().Z());
    printf("  Fov = %g %g\n", camera.XFOV(), camera.YFOV());
    printf("  Near = %g\n", camera.Near());
    printf("  Far = %g\n", camera.Far());
    fflush(stdout);
  }

  // Return viewer
  return viewer;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if ((argc == 2) && (*argv[1] == '-')) {
    printf("Usage: symsam [options]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = 1; 
      else if (!strcmp(*argv, "-debug")) print_debug = 1; 
      else if (!strcmp(*argv, "-images")) make_images = 1;
      else if (!strcmp(*argv, "-show_points")) show_points = 1; 
      else if (!strcmp(*argv, "-show_axes")) show_axes = 1; 
      else if (!strcmp(*argv, "-show_threshold")) {
        argc--; argv++; show_threshold = atof(*argv); 
      }
      else if (!strcmp(*argv, "-camera")) { 
        RNCoord x, y, z, tx, ty, tz, ux, uy, uz;
        argv++; argc--; x = atof(*argv); 
        argv++; argc--; y = atof(*argv); 
        argv++; argc--; z = atof(*argv); 
        argv++; argc--; tx = atof(*argv); 
        argv++; argc--; ty = atof(*argv); 
        argv++; argc--; tz = atof(*argv); 
        argv++; argc--; ux = atof(*argv); 
        argv++; argc--; uy = atof(*argv); 
        argv++; argc--; uz = atof(*argv); 
        initial_camera_origin = R3Point(x, y, z);
        initial_camera_towards.Reset(tx, ty, tz);
        initial_camera_up.Reset(ux, uy, uz);
        initial_camera = TRUE;
      }
      else if (!strcmp(*argv, "-background")) { 
        argc--; argv++; background_color[0] = atof(*argv); 
        argc--; argv++; background_color[1] = atof(*argv); 
        argc--; argv++; background_color[2] = atof(*argv); 
      }
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        exit(1); 
      }
      argv++; argc--;
    }
    else {
      if (!map_name) map_name = *argv;
      else { 
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        exit(1); 
      }
      argv++; argc--;
    }
  }

  // Check map filename
  if (!map_name) {
    fprintf(stderr, "You did not specify a map file.\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Initialize GLUI
  GLUIInit(&argc, argv);

  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read map
  map = ReadMap(map_name);
  if (!map) exit(-1);

  // Create viewer
  viewer = CreateViewer(map);
  if (!viewer) exit(-1);

  // Run GLUT interface
  GLUIMainLoop();

  // Return success 
  return 0;
}



