// Source file for the program that converts an ascii list of numbers to a binary one


// Include files

#include "RNBasics/RNBasics.h"



// Program arguments

static char *txt_name = 0;
static char *hst_name = 0;
static int print_verbose = 0;



static RNArray<char *> *
ReadText(const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = stdin;
  if (filename) {
    fp = fopen(filename, "r");
    if (!fp) {
      fprintf(stderr, "Unable to open text file: %s\n", filename);
      return 0;
    }
  }

  // Allocate histogram
  RNArray<char *> *histogram = new RNArray<char *>();
  if (!histogram) {
    fprintf(stderr, "Unable to allocate histogram\n");
    return 0;
  }

  // Read all lines from text file
  char buffer[1024];
  while (fscanf(fp, "%s", buffer) == 1) {
    char *str = strdup(buffer);
    histogram->Insert(str);
  }

  // Close file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Read text ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Values = %d\n", histogram->NEntries());
    fflush(stdout);
  }

  // Return histogram
  return histogram;
}



static int
WriteHistogram(RNArray<char *> *histogram, const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open file
  FILE *fp = stdout;
  if (filename) {
    fp = fopen(filename, "wb");
    if (!fp) {
      fprintf(stderr, "Unable to open histogram file %s\n", filename);
      return 0;
    }
  }

  // Write number of values
  int nvalues = histogram->NEntries();
  if (fwrite(&nvalues, sizeof(int), 1, fp) != 1) {
    fprintf(stderr, "Unable to write number of values to histogram file %s\n", filename);
    return 0;
  }

  // Write values
  for (int i = 0; i < nvalues; i++) {
    char *str = histogram->Kth(i);
    float value = atof(str);
    if (fwrite(&value, sizeof(float), 1, fp) != 1) {
      fprintf(stderr, "Unable to write value to histogram file %s\n", filename);
      return 0;
    }
  }

  // Close file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Wrote histogram ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", sizeof(int) + nvalues * sizeof(float));
    fflush(stdout);
  }

  // Return number of values written
  return nvalues;
}



static int 
ParseArgs(int argc, char **argv)
{
  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = 1; 
      else { fprintf(stderr, "Invalid program argument: %s", *argv); exit(1); }
    }
    else {
      if (!txt_name) txt_name = *argv;
      else if (!hst_name) hst_name = *argv;
      else { fprintf(stderr, "Invalid program argument: %s\n", *argv); exit(1); }
    }
    argv++; argc--;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read grid file
  RNArray<char *> *histogram = ReadText(txt_name);
  if (!histogram) exit(-1);

  // Write histogram
  int status = WriteHistogram(histogram, hst_name);
  if (!status) exit(-1);

  // Return success
  return 0;
}
